import bz2
import lzma
import zlib
from collections.abc import Callable

__all__ = ["ARCHIVERS", "UNARCHIVERS", "Archiver"]


type Archiver = Callable[[bytes], bytes]


def compress_zlib(data: bytes) -> bytes:
    return zlib.compress(data, level=9)


def compress_bz2(data: bytes) -> bytes:
    return bz2.compress(data, compresslevel=9)


def compress_lzma(data: bytes) -> bytes:
    return lzma.compress(
        data,
        format=lzma.FORMAT_ALONE,
        check=lzma.CHECK_NONE,
        preset=lzma.PRESET_EXTREME,
    )


def decompress_zlib(data: bytes) -> bytes:
    return zlib.decompress(data)


def decompress_bz2(data: bytes) -> bytes:
    return bz2.decompress(data)


def decompress_lzma(data: bytes) -> bytes:
    return lzma.decompress(data, format=lzma.FORMAT_ALONE)


def noop(data: bytes) -> bytes:
    return data


ARCHIVERS: dict[str, Archiver] = {
    "zlib": compress_zlib,
    "bz2": compress_bz2,
    "lzma": compress_lzma,
    "raw": noop,
}

UNARCHIVERS: dict[str, Archiver] = {
    "zlib": decompress_zlib,
    "bz2": decompress_bz2,
    "lzma": decompress_lzma,
    "raw": noop,
}



from nacl.public import Box, PrivateKey, PublicKey
from nacl.utils import EncryptedMessage

from rtty_soda.encoders import Encoder

__all__ = ["decrypt", "encrypt"]


def encrypt(
    private: PrivateKey, public: PublicKey, data: bytes, out_enc: Encoder
) -> EncryptedMessage:
    box = Box(private, public)
    return box.encrypt(data, encoder=out_enc)


def decrypt(
    private: PrivateKey, public: PublicKey, data: bytes, in_enc: Encoder
) -> bytes:
    box = Box(private, public)
    return box.decrypt(data, encoder=in_enc)



# ruff: noqa: PLR0913 - not applicable because we don't call Click Commands directly
# ruff: noqa: ANN201 - not applicable because the Click Commands are untyped
# mypy: disable-error-code = "no-untyped-def, misc"
from pathlib import Path

import click
from click_aliases import ClickAliasedGroup
from nacl.public import PrivateKey, PublicKey

from rtty_soda.archivers import ARCHIVERS, UNARCHIVERS
from rtty_soda.cli_io import (
    print_stats,
    read_bytes,
    read_ciphertext_bytes,
    read_clean_bytes,
    read_plaintext_bytes,
    write_output,
)
from rtty_soda.cryptography import public, secret
from rtty_soda.cryptography.kdf import KDF_PROFILES, kdf
from rtty_soda.encoders import ENCODERS, RawEncoder

in_path = click.Path(
    exists=True,
    file_okay=True,
    dir_okay=False,
    readable=True,
    allow_dash=True,
    path_type=Path,
)

out_path = click.Path(
    file_okay=True, dir_okay=False, writable=True, allow_dash=True, path_type=Path
)

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(cls=ClickAliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(package_name="rtty-soda")
def cli():
    pass


@cli.command()
@click.option("--encoding", "-e", default="base64", show_default=True)
def genkey_cmd(encoding: str):
    """Generate Private Key.

    Encoding: base26 | base36 | base64 | base94
    """
    enc = ENCODERS[encoding]
    key = PrivateKey.generate()
    click.echo(key.encode(enc))


@cli.command()
@click.argument("private_key_file", type=in_path)
@click.option("--encoding", "-e", default="base64", show_default=True)
def pubkey_cmd(private_key_file: Path, encoding: str):
    """Get Public Key.

    Encoding: base26 | base36 | base64 | base94
    """
    enc = ENCODERS[encoding]
    key = read_clean_bytes(private_key_file)
    priv = PrivateKey(key, enc)
    click.echo(priv.public_key.encode(enc))


@cli.command()
@click.argument("password_file", type=in_path)
@click.option("--encoding", "-e", default="base64", show_default=True)
@click.option("--profile", "-p", default="sensitive", show_default=True)
def kdf_cmd(password_file: Path, encoding: str, profile: str):
    """Key Derivation Function.

    Encoding: base26 | base36 | base64 | base94

    Profile: interactive | moderate | sensitive
    """
    enc = ENCODERS[encoding]
    prof = KDF_PROFILES[profile]
    pw = read_bytes(password_file)
    key = kdf(pw, prof, enc)
    click.echo(key)


@cli.command(aliases=["e"])
@click.argument("private_key_file", type=in_path)
@click.argument("public_key_file", type=in_path)
@click.argument("message_file", type=in_path)
@click.option("--key-encoding", default="base64", show_default=True)
@click.option("--data-encoding", "-e", default="base64", show_default=True)
@click.option("--compression", "-c", default="zlib", show_default=True)
@click.option("--output-file", "-o", type=out_path, help="(Optional)")
def encrypt_public_cmd(
    private_key_file: Path,
    public_key_file: Path,
    message_file: Path,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
):
    """Encrypt Message (Public).

    Encoding: base26 | base36 | base64 | base94 | binary

    Compression: zlib | bz2 | lzma | raw
    """
    key_enc = ENCODERS[key_encoding]
    data_enc = ENCODERS[data_encoding]
    archiver = ARCHIVERS[compression]
    priv_bytes = read_clean_bytes(private_key_file)
    priv = PrivateKey(priv_bytes, key_enc)
    pub_bytes = read_clean_bytes(public_key_file)
    pub = PublicKey(pub_bytes, key_enc)
    data = stats = read_plaintext_bytes(message_file, data_enc)
    data = archiver(data)
    data = public.encrypt(private=priv, public=pub, data=data, out_enc=data_enc)
    write_output(output_file, data, data_enc)
    print_stats(stats, data)


@cli.command(aliases=["es"])
@click.argument("key_file", type=in_path)
@click.argument("message_file", type=in_path)
@click.option("--key-encoding", default="base64", show_default=True)
@click.option("--data-encoding", "-e", default="base64", show_default=True)
@click.option("--compression", "-c", default="zlib", show_default=True)
@click.option("--output-file", "-o", type=out_path, help="(Optional)")
def encrypt_secret_cmd(
    key_file: Path,
    message_file: Path,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
):
    """Encrypt Message (Secret).

    Encoding: base26 | base36 | base64 | base94 | binary

    Compression: zlib | bz2 | lzma | raw
    """
    key_enc = ENCODERS[key_encoding]
    data_enc = ENCODERS[data_encoding]
    archiver = ARCHIVERS[compression]
    key = read_clean_bytes(key_file)
    data = stats = read_plaintext_bytes(message_file, data_enc)
    data = archiver(data)
    data = secret.encrypt(key=key, data=data, key_enc=key_enc, out_enc=data_enc)
    write_output(output_file, data, data_enc)
    print_stats(stats, data)


@cli.command(aliases=["ep"])
@click.argument("password_file", type=in_path)
@click.argument("message_file", type=in_path)
@click.option("--kdf-profile", "-p", default="sensitive", show_default=True)
@click.option("--data-encoding", "-e", default="base64", show_default=True)
@click.option("--compression", "-c", default="zlib", show_default=True)
@click.option("--output-file", "-o", type=out_path, help="(Optional)")
def encrypt_password_cmd(
    password_file: Path,
    message_file: Path,
    kdf_profile: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
):
    """Encrypt Message (Password).

    KDF profile: interactive | moderate | sensitive

    Encoding: base26 | base36 | base64 | base94 | binary

    Compression: zlib | bz2 | lzma | raw
    """
    prof = KDF_PROFILES[kdf_profile]
    data_enc = ENCODERS[data_encoding]
    archiver = ARCHIVERS[compression]
    pw = read_bytes(password_file)
    key = kdf(pw, prof, RawEncoder)
    data = stats = read_plaintext_bytes(message_file, data_enc)
    data = archiver(data)
    data = secret.encrypt(key=key, data=data, key_enc=RawEncoder, out_enc=data_enc)
    write_output(output_file, data, data_enc)
    print_stats(stats, data)


@cli.command(aliases=["d"])
@click.argument("private_key_file", type=in_path)
@click.argument("public_key_file", type=in_path)
@click.argument("message_file", type=in_path)
@click.option("--key-encoding", default="base64", show_default=True)
@click.option("--data-encoding", "-e", default="base64", show_default=True)
@click.option("--compression", "-c", default="zlib", show_default=True)
@click.option("--output-file", "-o", type=out_path, help="(Optional)")
def decrypt_public_cmd(
    private_key_file: Path,
    public_key_file: Path,
    message_file: Path,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
):
    """Decrypt Message (Public).

    Encoding: base26 | base36 | base64 | base94 | binary

    Compression: zlib | bz2 | lzma | raw
    """
    key_enc = ENCODERS[key_encoding]
    data_enc = ENCODERS[data_encoding]
    unarchiver = UNARCHIVERS[compression]
    priv_bytes = read_clean_bytes(private_key_file)
    priv = PrivateKey(priv_bytes, key_enc)
    pub_bytes = read_clean_bytes(public_key_file)
    pub = PublicKey(pub_bytes, key_enc)
    data = stats = read_ciphertext_bytes(message_file, data_enc)
    data = public.decrypt(priv, pub, data, data_enc)
    data = unarchiver(data)
    write_output(output_file, data, data_enc)
    print_stats(data, stats)


@cli.command(aliases=["ds"])
@click.argument("key_file", type=in_path)
@click.argument("message_file", type=in_path)
@click.option("--key-encoding", default="base64", show_default=True)
@click.option("--data-encoding", "-e", default="base64", show_default=True)
@click.option("--compression", "-c", default="zlib", show_default=True)
@click.option("--output-file", "-o", type=out_path, help="(Optional)")
def decrypt_secret_cmd(
    key_file: Path,
    message_file: Path,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
):
    """Decrypt Message (Secret).

    Encoding: base26 | base36 | base64 | base94 | binary

    Compression: zlib | bz2 | lzma | raw
    """
    key_enc = ENCODERS[key_encoding]
    data_enc = ENCODERS[data_encoding]
    unarchiver = UNARCHIVERS[compression]
    key = read_clean_bytes(key_file)
    data = stats = read_ciphertext_bytes(message_file, data_enc)
    data = secret.decrypt(key, data, key_enc, data_enc)
    data = unarchiver(data)
    write_output(output_file, data, data_enc)
    print_stats(data, stats)


@cli.command(aliases=["dp"])
@click.argument("password_file", type=in_path)
@click.argument("message_file", type=in_path)
@click.option("--kdf-profile", "-p", default="sensitive", show_default=True)
@click.option("--data-encoding", "-e", default="base64", show_default=True)
@click.option("--compression", "-c", default="zlib", show_default=True)
@click.option("--output-file", "-o", type=out_path, help="(Optional)")
def decrypt_password_cmd(
    password_file: Path,
    message_file: Path,
    kdf_profile: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
):
    """Decrypt Message (Password).

    KDF profile: interactive | moderate | sensitive

    Encoding: base26 | base36 | base64 | base94 | binary

    Compression: zlib | bz2 | lzma | raw
    """
    prof = KDF_PROFILES[kdf_profile]
    data_enc = ENCODERS[data_encoding]
    unarchiver = UNARCHIVERS[compression]
    pw = read_bytes(password_file)
    key = kdf(pw, prof, RawEncoder)
    data = stats = read_ciphertext_bytes(message_file, data_enc)
    data = secret.decrypt(key, data, RawEncoder, data_enc)
    data = unarchiver(data)
    write_output(output_file, data, data_enc)
    print_stats(data, stats)


@cli.command()
@click.argument("in_encoding")
@click.argument("out_encoding")
@click.argument("file", type=in_path)
@click.option("--output-file", "-o", type=out_path, help="(Optional)")
def encode_cmd(
    in_encoding: str, out_encoding: str, file: Path, output_file: Path | None
):
    """Encode File.

    Encoding: base26 | base36 | base64 | base94 | binary
    """
    in_enc = ENCODERS[in_encoding]
    out_enc = ENCODERS[out_encoding]
    data = read_ciphertext_bytes(file, in_enc)
    data = in_enc.decode(data)
    data = out_enc.encode(data)
    write_output(output_file, data, out_enc)


if __name__ == "__main__":
    cli()



import random
import re
import string
from pathlib import Path
from typing import TextIO, cast

import click

from rtty_soda.encoders import Encoder, RawEncoder, encode_str

__all__ = [
    "print_stats",
    "read_bytes",
    "read_ciphertext_bytes",
    "read_clean_bytes",
    "read_plaintext_bytes",
    "read_str",
    "remove_whitespace",
    "write_bytes_atomic",
    "write_output",
]


def read_str(source: Path) -> str:
    with click.open_file(source, mode="rt", encoding="utf-8", errors="strict") as fd:
        return cast("TextIO", fd).read().strip()


def read_bytes(source: Path) -> bytes:
    return encode_str(read_str(source))


def remove_whitespace(data: str) -> str:
    return re.sub(r"\s", "", data)


def read_clean_bytes(source: Path) -> bytes:
    data = read_str(source)
    data = remove_whitespace(data)
    return encode_str(data)


def read_plaintext_bytes(source: Path, in_enc: Encoder) -> bytes:
    if in_enc == RawEncoder:
        return source.read_bytes()

    return read_bytes(source)


def read_ciphertext_bytes(source: Path, in_enc: Encoder) -> bytes:
    if in_enc == RawEncoder:
        return source.read_bytes()

    return read_clean_bytes(source)


def write_bytes_atomic(target: Path, data: bytes) -> None:
    temp_name = "".join(random.choices(string.ascii_lowercase, k=10))  # noqa: S311
    temp_path = target.parent / temp_name
    temp_path.write_bytes(data)
    temp_path.replace(target)


def write_output(target: Path | None, data: bytes, out_enc: Encoder) -> None:
    if target is None or target.stem == "-":
        if out_enc == RawEncoder:
            click.confirm(
                "Print binary output to the terminal?", default=False, abort=True
            )

        click.echo()
        click.echo(data)
        click.echo()
    else:
        if target.exists():
            click.confirm(
                f"Overwrite the output file? ({target})", default=False, abort=True
            )

        write_bytes_atomic(target, data)


def print_stats(plaintext: bytes, ciphertext: bytes) -> None:
    click.echo(f"Plaintext: {len(plaintext)}", err=True)
    click.echo(f"Ciphertext: {len(ciphertext)}", err=True)
    overhead = len(ciphertext) / len(plaintext)
    click.echo(f"Overhead: {overhead:.3f}", err=True)



from nacl.encoding import Base64Encoder, Encoder, RawEncoder

from .base26_encoder import Base26Encoder
from .base36_encoder import Base36Encoder
from .base94_encoder import Base94Encoder
from .functions import decode_bytes, encode_str

__all__ = [
    "ENCODERS",
    "Base26Encoder",
    "Base36Encoder",
    "Base64Encoder",
    "Base94Encoder",
    "Encoder",
    "RawEncoder",
    "decode_bytes",
    "encode_str",
]

ENCODERS: dict[str, Encoder] = {
    "base26": Base26Encoder,
    "base36": Base36Encoder,
    "base64": Base64Encoder,
    "base94": Base94Encoder,
    "binary": RawEncoder,
}
